"""Test the version comparison routines."""

import unittest

from typing import NamedTuple

import ddt  # type: ignore

import trivver


VersionSet = NamedTuple(  # pylint: disable=invalid-name
    "VersionSet", [("first", str), ("second", str), ("expected", int),]
)


VERSIONS = [
    VersionSet(first="1.0", second="2.0", expected=-1),
    VersionSet(first="1.0", second="1.0.1", expected=-1),
    VersionSet(first="1.0", second="1.0a", expected=-1),
    VersionSet(first="1.0", second="1.0", expected=0),
    VersionSet(first="1.0a", second="1.0a", expected=0),
    VersionSet(first="0.1.0.b", second="0.1.0", expected=-1),
    VersionSet(
        first="3.10.0-1062.1.1.el7.x86_64",
        second="3.10.0-983.13.1.el7.x86_64",
        expected=1,
    ),
    VersionSet(
        first="3.10.0-1062.1.1.el7.x86_64",
        second="3.10.0-1062.1.1.el6.x86_64",
        expected=1,
    ),
]

UNSORTED = [
    "3.10.0-1062.1.1.el7.x86_64",
    "1.0a",
    "0.1.0.b",
    "1.0",
    "3.10.0-983.13.1.el7.x86_64",
    "0.1.0",
    "1.0.1",
    "2.0",
    "3.10.0-1062.1.1.el6.x86_64",
]
SORTED = [
    "0.1.0.b",
    "0.1.0",
    "1.0",
    "1.0.1",
    "1.0a",
    "2.0",
    "3.10.0-983.13.1.el7.x86_64",
    "3.10.0-1062.1.1.el6.x86_64",
    "3.10.0-1062.1.1.el7.x86_64",
]


def sign(value):
    # type: (int) -> int
    """Return the sign of an integer."""
    if value > 0:
        return 1
    if value < 0:
        return -1
    return 0


@ddt.ddt
class TestVersions(unittest.TestCase):
    """Test the version comparison routines."""

    # pylint: disable=no-self-use

    @ddt.data(*VERSIONS)
    @ddt.unpack
    def test_compare(self, first, second, expected):
        # type: (TestVersions, str, str, int) -> None
        """Compare a single pair of versions."""
        assert sign(trivver.compare(first, second)) == sign(expected)
        assert sign(
            trivver.compare(  # pylint: disable=arguments-out-of-order
                second, first
            )
        ) == sign(-expected)

    def test_sort(self):
        # type: (TestVersions) -> None
        """Sort a list of versions."""
        assert sorted(UNSORTED, key=trivver.key_compare) == SORTED



"""Compare package versions in all their varied glory.

This module provides the `compare()` function which compares two
version strings and returns a negative value, zero, or a positive
value depending on whether the first string represents a version
number lower than, equal to, or higher than the second one, and
the `key_compare()` function which may be used as a key for e.g.
`sorted()`.

This module does not strive for completeness in the formats of
version strings that it supports. Some version strings sorted by
its rules are:
- 0.1.0
- 0.2.alpha
- 0.2
- 0.2.1
- 0.2a
- 0.2a.1
- 0.2p3
- 1.0.beta
- 1.0.beta.2
- 1.0
"""

import functools
import re

from typing import List, Optional, Text, Union


RE_COMPONENT = re.compile(
    r""" ^
    (?P<num> [0-9]+ )? (?P<rest> .+ )?
$ """,
    re.X,
)

VERSION = "0.1.0"


def cmp_strings(first, second):
    # type: (Optional[Text], Optional[Text]) -> int
    """Compare two strings, possibly undefined."""
    if first is not None:
        if second is not None:
            if first == second:
                return 0
            if first > second:
                return 1
            return -1

        return 1

    if second is not None:
        return -1

    return 0


def cmp_components(first, second):
    # type: (Text, Text) -> int
    """Compare a single pair of version components.

    - alpha4 < beta
    - alpha4 < 1
    - 1 < 1sp2
    - 1 < 2"""
    a_m = RE_COMPONENT.match(first)
    assert a_m
    a_num = a_m.groupdict()["num"]
    a_rest = a_m.groupdict()["rest"]

    b_m = RE_COMPONENT.match(second)
    assert b_m
    b_num = b_m.groupdict()["num"]
    b_rest = b_m.groupdict()["rest"]

    if a_num is not None:
        if b_num is not None:
            if a_num != b_num:
                return int(a_num) - int(b_num)
            return cmp_strings(a_rest, b_rest)

        return 1

    if b_num is not None:
        return -1

    return cmp_strings(a_rest, b_rest)


def cmp_examine_extra(a_first):
    # type: (Text) -> int
    """Does the version component start with a digit?

    In other words, is it an alpha/beta version or not?"""
    a_m = RE_COMPONENT.match(a_first)
    assert a_m
    if a_m.groupdict()["num"] is not None:
        return 1
    return -1


def cmp_extra(first, second):
    # type: (List[Text], List[Text]) -> int
    """Check for alpha versions in the longer list's additional components.

    Does one of the lists:
    - have more components than the other, and
    - the first extra component starts with a digit, not a letter?
    (if it starts with a letter, it's an alpha/beta/pre-release version)"""
    a_len = len(first)
    b_len = len(second)
    if a_len == b_len:
        return 0

    if a_len > b_len:
        a_first = first[b_len]
        return cmp_examine_extra(a_first)

    b_first = second[a_len]
    return -cmp_examine_extra(b_first)


def cmp_dot_components(first, second):
    # type: (Text, Text) -> int
    """Split versions number no dots, compare them component by component."""
    a_dot = first.split(".")
    b_dot = second.split(".")
    for a_comp, b_comp in zip(a_dot, b_dot):
        res = cmp_components(a_comp, b_comp)
        if res != 0:
            return res

    # Does one of them have more dot-delimited components than the other?
    return cmp_extra(a_dot, b_dot)


# Split a version number on dashes, compare them component by component
def compare(first, second):
    # type: (Union[Text, str], Union[Text, str]) -> int
    """Compare two full versions strings."""
    a_dash = Text(first).split("-")
    b_dash = Text(second).split("-")
    for a_comp, b_comp in zip(a_dash, b_dash):
        res = cmp_dot_components(a_comp, b_comp)
        if res != 0:
            return res

    # Does one of them have more dash-delimited components than the other?
    # (don't check for beta versions in this case)
    return len(a_dash) - len(b_dash)


key_compare = functools.cmp_to_key(compare)  # pylint: disable=invalid-name

__all__ = ("VERSION", "compare", "key_compare")



#!/usr/bin/env python
#
# Copyright (c) 2020  Peter Pentchev <roam@ringlet.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Setup configuration for the trivial version comparison module."""

import io
import re

from typing import Text, cast

import setuptools  # type: ignore


RE_VERSION = re.compile(
    r"""^
    \s* VERSION \s* = \s* "
    (?P<version>
           (?: 0 | [1-9][0-9]* )    # major
        \. (?: 0 | [1-9][0-9]* )    # minor
        \. (?: 0 | [1-9][0-9]* )    # patchlevel
    (?: \. [a-zA-Z0-9]+ )?          # optional addendum (dev1, beta3, etc.)
    )
    " \s*
    $""",
    re.X,
)


def get_version():
    # type: () -> Text
    """ Get the version string from the module's __init__ file. """
    with io.open("trivver/__init__.py", encoding="UTF-8") as init:
        found = [
            match.group("version")
            for match in (RE_VERSION.match(cast(Text, line)) for line in init)
            if match
        ]

    assert len(found) == 1
    return found[0]


def get_long_description():
    # type: () -> Text
    """Read the long description from the README file."""
    with io.open("README.md", mode="r", encoding="UTF-8") as readme:
        return cast(Text, readme.read())


setuptools.setup(
    name="trivver",
    version=get_version(),
    packages=("trivver",),
    author="Peter Pentchev",
    author_email="roam@ringlet.net",
    description="A library for comparing version strings",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    license="Apache License 2.0",
    keywords="version",
    url="https://gitlab.com/ppentchev/python-trivver",
    setup_requires=["typing"],
    install_requires=["typing"],
    package_data={"trivver": ["py.typed"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: DFSG approved",
        "License :: Freely Distributable",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    zip_safe=True,
)
